#include "lie.h"


static void simp_rt_refl (entry* w,index i,simpgrp* g);
static void rt_refl (entry* w,index i);
static void simp_rt_refls (entry** m,index offset,index n,index i,simpgrp* g);


/* optimised routines that bypass the cartan matrix */

/* simple reflection i on weight basis */
void simp_w_refl(entry* w, index i, simpgrp* g)
{ index r=g->lierank; entry wi=w[i];
  if (!wi) return;
  if (i>0) w[i-1]+=wi;	w[i]= -wi;  if (i<r-1) w[i+1]+=wi;
  switch (g->lietype)
  {	 case 'B': if (i==r-2) w[i+1]+=wi;
  break; case 'C': if (i==r-1) w[i-1]+=wi;
  break; case 'D': if (i>=r-3) if (i==r-1) { w[i-2]+=wi; w[i-1]-=wi; }
			       else if (i==r-2) w[i+1]-=wi; else w[i+2]+=wi;
  break; case 'E': if (i<4)
    { if (i<2) { w[i+1]-=wi; w[i+2]+=wi; } else w[i-2]+=wi;
      if (i==1 || i==2) w[i-1]-=wi;
    }
  break; case 'F': if (i==1) w[2]+=wi;
  break; case 'G': if (i==1) w[0]+=2*wi;
  }
}

void w_refl(entry* w, index nr)
{ if (type_of(grp)==SIMPGRP) simp_w_refl(w,nr,&grp->s);
  else if (simpgroup(grp)) simp_w_refl(w,nr,Liecomp(grp,0));
  else
  { index i,d,offset=0;
    for (i=0; nr>=(d=Liecomp(grp,i)->lierank); ++i) { offset+=d; nr-=d; }
    simp_w_refl(w+offset,nr,Liecomp(grp,i));
  }
}

/* Waction(lambda,word) applies the simple relflections encoded in
   `word' from left to right to the weight vector `lambda'.
*/
void Waction(entry* lambda, vector* word)
{ index i; entry* w=word->compon;
  for (i=0; i<word->ncomp; i++) if (w[i]) w_refl(lambda,w[i]-1);
}

/* This routine is large because we have taken the tests out of the loops */
#define loop  while(n-->0 && (v=(*m++)+k,true))   /* this helps a bit */

static void simp_w_refls (entry** m,index offset,index n,index i,simpgrp* g)
{ index k=offset+i,r=g->lierank; entry* v;
  switch (g->lietype)
  {	 case 'A':
    if (i>0 && i<r-1)	loop { v[-1]+= *v; v[1]+= *v; *v= -*v; }
    else if (i==0) if (r==1) loop		      *v= -*v;
		   else loop {		   v[1]+= *v; *v= -*v; }
    else /* i==r-1 */	loop { v[-1]+= *v;	      *v= -*v; }
  break; case 'B':
    if (i>0 && i<r-2) loop { v[-1]+= *v; v[1]+= *v;    *v= -*v; }
    else if (i==0) if (r==2) loop {	 v[1]+=2*(*v); *v= -*v; }
		   else loop {		 v[1]+= *v;    *v= -*v; }
    else if (i==r-1)  loop { v[-1]+= *v;	       *v= -*v; }
    else /* i==r-2 */ loop { v[-1]+= *v; v[1]+=2*(*v); *v= -*v; }
  break; case 'C':
    if (i>0 && i<r-1) loop { v[-1]+= *v; v[1]+= *v; *v= -*v; }
    else if (i==0)    loop {		 v[1]+= *v; *v= -*v; }
    else /* i==r-1 */ loop { v[-1]+=2*(*v);	    *v= -*v; }
  break; case 'D':
    if (i>0 && i<r-3)	     loop { v[-1]+= *v; v[1]+= *v;	      *v= -*v; }
    else if (i==0) if (r==3) loop {		v[1]+= *v; v[2]+= *v; *v= -*v; }
		   else      loop {		v[1]+= *v;	       *v= -*v; }
    else if (i==r-3)	     loop { v[-1]+= *v; v[1]+= *v; v[2]+= *v; *v= -*v; }
    else /* i==r-2 or i==r-1 */ { index j=r-3-i; loop { v[j]+= *v; *v= -*v; }}
  break; case 'E':
    if (i>3 && i<r-1) loop { v[-1]+= *v; v[1]+= *v; *v= -*v; }
    else if (i==r-1)  loop { v[-1]+= *v;	    *v= -*v; }
    else if (i<2) loop { v[2]+= *v; *v= -*v; } 
    else if (i==2)  loop { v[-2]+= *v;		   v[1]+= *v; *v= -*v; }
    else /* i==3 */ loop { v[-2]+= *v; v[-1]+= *v; v[1]+= *v; *v= -*v; }
  break; case 'F': switch(i)
    {	   case 0: loop {	      v[1]+= *v;    *v= -*v; }
    break; case 1: loop { v[-1]+= *v; v[1]+=2*(*v); *v= -*v; }
    break; case 2: loop { v[-1]+= *v; v[1]+= *v;    *v= -*v; }
    break; case 3: loop { v[-1]+= *v;		    *v= -*v; }
    }
  break; case 'G':
    if (i==0) loop { v[1]+= *v; *v= -*v; } else loop { v[-1]+=3*(*v); *v= -*v; }
  }
}
#undef loop

void w_refls(m,n,nr) entry** m; index n,nr;
{ if (type_of(grp)==SIMPGRP) simp_w_refls(m,0,n,nr,&grp->s);
  else if (simpgroup(grp)) simp_w_refls(m,0,n,nr,Liecomp(grp,0));
  else
  { index i,d,offset=0;
    for (i=0; nr>=(d=Liecomp(grp,i)->lierank); i++) { offset+=d; nr-=d; }
    simp_w_refls(m,offset,n,nr,Liecomp(grp,i));
  }
}

/* simple reflection i on root basis */
static void simp_rt_refl(w,i,g) entry* w; index i; simpgrp* g;
{ index r=g->lierank; entry c= -w[i]+(i>0?w[i-1]:0)+(i<r-1?w[i+1]:0);
  switch (g->lietype)
  {	 case 'B': if (i==r-1) c+=w[i-1];
  break; case 'C': if (i==r-2) c+=w[i-2];
  break; case 'D': if (i>=r-3) if (i==r-1) c+=w[i-2]-w[i-1];
			       else if (i==r-2) c-=w[i+1]; else c+=w[i+2];
  break; case 'E': if (i<4)
    { if (i<2) c+=w[i+2]-w[i+1]; else c+=w[i-2]; if (i==1 || i==2) c-=w[i-1]; }
  break; case 'F': if (i==2) c+=w[1];
  break; case 'G': if (i==0) c+=2*w[1];
  }
  w[i]=c;
}

static void rt_refl(w,nr) entry* w; index nr;
{ if (type_of(grp)==SIMPGRP) simp_rt_refl(w,nr,&grp->s);
  else if (simpgroup(grp)) simp_rt_refl(w,nr,Liecomp(grp,0));
  else
  { index i,d,offset=0;
    for (i=0; nr>=(d=Liecomp(grp,i)->lierank); i++) { offset+=d; nr-=d; }
    simp_rt_refl(w+offset,nr,Liecomp(grp,i));
  }
}

/* Wrtaction(root,word) applies the simple relflections encoded in
   `word' from left to right to the root vector `root'.
*/
void Wrtaction(root,word) entry* root; vector* word;
{ index i; entry* w=word->compon;
  for (i=0; i<word->ncomp; i++) if(w[i]) rt_refl(root,w[i]-1); }

/* This routine is large because we have taken the tests out of the loops. */
#define loop  while(n-->0 && (v=(*m++)+k,true))   /* this helps a bit */

static void simp_rt_refls(m,offset,n,i,g)
  entry** m; index offset,n,i; simpgrp* g;
{ index k=offset+i,r=g->lierank; entry* v;
  switch (g->lietype)
  {	 case 'A':
    if (i>0 && i<r-1)	     loop *v=v[-1]-*v+v[1];
    else if (i==0) if (r==1) loop *v=	  -*v;
		   else      loop *v=	  -*v+v[1];
    else /* i==r-1 */	     loop *v=v[-1]-*v;
  break; case 'B':
    if (i>0 && i<r-1) loop *v=	v[-1]-*v+v[1];
    else if (i==0)    loop *v=	     -*v+v[1];
	 else	      loop *v=2*v[-1]-*v;
  break; case 'C':
    if (i>0 && i<r-2)	     loop *v=v[-1]-*v+v[1];
    else if (i==0) if (r==2) loop *v=	  -*v+2*v[1]; 
		   else      loop *v=	  -*v+v[1];
    else if (i==r-1)	     loop *v=v[-1]-*v;
    else /* i==r-2 */	     loop *v=v[-1]-*v+2*v[1];
  break; case 'D':
    if (i>0 && i<r-3)	     loop *v=v[-1]-*v+v[1];
    else if (i==0) if (r==3) loop *v=	  -*v+v[1]+v[2];
		   else      loop *v=	  -*v+v[1];
    else if (i==r-3)	     loop *v=v[-1]-*v+v[1]+v[2];
    else /* i==r-2 or i==r-1 */ { index j=r-3-i; loop *v= -*v+v[j]; }
  break; case 'E':
    if (i>3 && i<r-1) loop *v=v[-1]-*v+v[1];
    else if (i==r-1) loop *v=v[-1]-*v;
    else if (i<2) loop *v= -*v+v[2];
    else if (i==2) loop *v=v[-2]-*v+v[1];
    else loop *v=v[-2]+v[-1]-*v+v[1];
  break; case 'F': switch(i)
    {	   case 0: loop *v= -*v+v[1];
    break; case 1: loop *v=v[-1]-*v+v[1];
    break; case 2: loop *v=2*v[-1]-*v+v[1];
    break; case 3: loop *v=v[-1]-*v;
    }
  break; case 'G':
    if (i==0) loop *v= -*v+3*v[1]; else loop *v=v[-1]-*v;
  }
}
#undef loop

void rt_refls(m,n,nr) entry** m; index n,nr;
{ if (type_of(grp)==SIMPGRP) simp_rt_refls(m,0,n,nr,&grp->s);
  else if (simpgroup(grp)) simp_rt_refls(m,0,n,nr,Liecomp(grp,0));
  else
  { index i,d,offset=0;
    for (i=0; nr>=(d=Liecomp(grp,i)->lierank); i++) { offset+=d; nr-=d; }
    simp_rt_refls(m,offset,n,nr,Liecomp(grp,i));
  }
}

matrix* simp_Weylmat(base,word,g) boolean base; vector* word; simpgrp* g;
{ index i,j,r=g->lierank; matrix* res=mkmatrix(r,r);
  entry** m=res->elm,* w=word->compon,wi;
  for (i=0; i<r; i++)  for (j=0; j<r; j++) m[i][j]= i==j;
  if (base)
  { for (i=0; i<word->ncomp; i++)
      if((wi=w[i])!=0) simp_rt_refls(m,0,r,wi-1,g);
  } /* braces needed to properly bind following else */
  else
    for (i=0; i<word->ncomp; i++)
      if((wi=w[i])!=0) simp_w_refls(m,0,r,wi-1,g);
  return res;
}

matrix* Weylmat(base,word) boolean base; vector* word;
{ index i,j,r= base ? Ssrank(grp) : Lierank(grp); matrix* res=mkmatrix(r,r);
  entry** m=res->elm,* w=word->compon,wi;
  for (i=0; i<r; i++)  for (j=0; j<r; j++) m[i][j]= i==j;
  if (base)
  { for (i=0; i<word->ncomp; i++) if((wi=w[i])!=0) rt_refls(m,r,wi-1);
  } /* braces needed to properly bind following else */
  else
    for (i=0; i<word->ncomp; i++) if((wi=w[i])!=0) w_refls(m,r,wi-1);
  return res;
}


bigint* simp_worder(num,g) bigint* num; simpgrp* g;
{ index i,r=g->lierank; bigint* result=num;

  i=r; while (i>1) result=mul1(result,i--);
  switch (g->lietype)
  {	 case 'A': result=mul1(result,r+1);
  break; case 'B':
	 case 'C': for (i=0; i<r; i++) result = mul1(result,2);
  break; case 'D': for (i=1; i<r; i++) result = mul1(result,2);
  break; case 'E': result = mul1(result,r==6 ? 72 : r==7 ? 576 : 17280);
  break; case 'F': result = mul1(result,48);
  break; case 'G': result = mul1(result,6);
  }
  return result;
}

bigint* Worder(grp) object grp;
{ index i; bigint* result=copybigint(one,NULL);
  if (type_of(grp)==SIMPGRP) return simp_worder(result,&grp->s);
  for (i=0; i<grp->g.ncomp; i++) result = simp_worder(result,Liecomp(grp,i));
  return result;
}

/* order of Weyl subgroup generated by given subset of simple reflections */
bigint* sub_Worder(v) vector* v;
{ index i,j,s=Ssrank(grp), n=v->ncomp; matrix* roots=mkmatrix(n,s);
  entry** m=roots->elm; group* h; bigint* result;
  if (n==0) { freemem(roots); return one; }
  for (i=0; i<n; i++)
  { entry* mij= *m++,vi=v->compon[i]-1;
    for (j=0; j<s; j++) *(mij++)= (j==vi);
  }
  h=Carttype(roots); freemem(roots);
  result= Worder((object)h); freemem(h); return(result);
}
