#include "lie.h"

static bigint* simp_stabsize(entry* v,simpgrp* g);
bigint* simp_worbitsize(entry* w,simpgrp* g); /* w assumed dominant */
static void addorbvec(entry* orb_v);
static matrix* simp_worbit(boolean base,entry* v,simpgrp* g);
/* base as in Weylloopinit: false for weight coords, true for root coords */


/* size of the stabiliser of a DOMINANT weight */
static bigint* simp_stabsize(entry* v, simpgrp* g)
{ object sav_grp=grp; vector* I; bigint* result;
  index i,nz=0,r=g->lierank;
  for (i=0; i<r; i++) if (v[i]==0) nz++;
  if (nz==0) return one;
  if (nz==r) { return simp_worder(copybigint(one,NULL),g); }
  I=mkvector(nz); for (i=0,nz=0; i<r; i++) if (v[i]==0) I->compon[nz++]=i+1;
  grp=(object)g; result=sub_Worder(I); grp=sav_grp;
  freemem(I); return result;
}

bigint* simp_worbitsize(entry* w, simpgrp* g) /* w assumed dominant */
{ return quotient(simp_worder(copybigint(one,NULL),g),simp_stabsize(w,g)); }

/* Compute size of orbit of vector of length `Lierank(grp)' on weight base */
bigint* Orbitsize(entry* w)
{ index i,d,s=Ssrank(grp); entry* x=mkintarray(s),* y=x; bigint* result=one;
  copyrow(w,x,s); make_dominant(x);
  if (type_of(grp)==SIMPGRP) return simp_worbitsize(x,&grp->s);
  for (i=0; i<grp->g.ncomp; i++,y+=d)
  { simpgrp* g=Liecomp(grp,i); d=g->lierank;
    result=mult(result,simp_worbitsize(y,g));
  }
  freearr(x); return result;
}

static matrix* orb_mat;
static index orb_i,rnk,orb_l;
static entry** orbit;

static void addorbvec(entry* orb_v)
{ if (orb_i>=orb_l) error("Error in orbit: too large\n");
  copyrow(orb_v,orbit[orb_i++],rnk);
}

static matrix* simp_worbit(boolean base, entry* v, simpgrp* g)
{ index r=g->lierank;
  { entry* w=mkintarray(r);
    if (base) mulvecmatelm(v,simp_Cartan(g)->elm,w,r,r); else copyrow(v,w,r);
    simp_make_dominant(w,g); orb_l= bigint2entry(simp_worbitsize(w,g));
    freearr(w);
  }
  orb_mat=mkmatrix(orb_l,rnk=r); orbit=orb_mat->elm; orb_i=0;
  Weylloopinit(base,g); Weylloop(addorbvec,v); Weylloopexit();
  if (orb_i!=orb_l)
  { error("Error in orbit: size %ld is too small \n",(long)orb_i); }
  return orb_mat;
}

matrix* Worbit(boolean base, entry* v)
{ index i,j,r=base ? Ssrank(grp) : Lierank(grp); entry* x=mkintarray(r);
  matrix* result; copyrow(v,x,r);
  if (type_of(grp)==SIMPGRP) result=simp_worbit(base,x,&grp->s);
  else if (simpgroup(grp)) result=simp_worbit(base,x,grp->g.liecomp[0]);
  else
  { index n=grp->g.ncomp; matrix** M=alloc_array(matrix*,n);
    for (i=0,j=0; i<n; j+=grp->g.liecomp[i++]->lierank)
      M[i]=simp_worbit(base,&x[j],grp->g.liecomp[i]);
    result=Cartes_prod(M,n,&x[j],r);
  }
  freearr(x); return result;
}

poly* Worbit_p(poly* p)
{ index i,k=0,r=p->ncols; poly* result; entry** res;
  p=copypoly(p); for (i=0; i<p->nrows; i++) make_dominant(p->elm[i]);
  Reduce_pol(p);
  for (i=0,orb_l=0; i<p->nrows; i++)
    if ((orb_l += bigint2entry(Orbitsize(p->elm[i])))<0)
      error ("That's too large an orbit");
  result=mkpoly(orb_l,p->ncols); res=result->elm;
  for (i=0; i<p->nrows; i++)
  { index j; matrix* orbit=Worbit(false,p->elm[i]); entry** x=orbit->elm;
    for (j=0; j<orbit->nrows; ++j)
    { result->coef[k]=p->coef[i]; setshared(p->coef[i]);
      copyrow(*x++,res[k++],r);
    }
    freemem(orbit);
  }
  if (k!=result->nrows)
  { error("Error in orbit: size %ld is too %s.\n"
	 , (long)k, k<result->nrows ? "small" : "large"); }
  return result; /* not sorted, but rows are unique */
}

poly* alt_Wsum(poly* p)
{ index i,k=0,r=p->ncols; poly* result; entry** res,*rho=mkintarray(r);
  p=Alt_dom(p); for (i=0; i<r; i++) rho[i]=1;
  for (i=0; i<p->nrows; i++) add_xrow_to(p->elm[i],1,rho,r);
  result=mkpoly(p->nrows*bigint2entry(Worder(grp)),r); res=result->elm;
  for (i=0; i<p->nrows; i++)
  { index j,l; matrix* orbit=Worbit(false,p->elm[i]); entry** x=orbit->elm;
    bigint* c=p->coef[i],* min_c=sub(null,c);
    for (j=0; j<orbit->nrows; j++)
    { subrow(*x,rho,res[k],r); l=make_dominant(*x++)%2;
      result->coef[k]= l ? min_c : c; setshared(result->coef[k]); k++;
    }
    freemem(orbit);
  }
  freearr(rho);
  if (k!=result->nrows)
  { error("Error in orbit: size %ld is too %s.\n"
	 , (long)k, k<result->nrows ? "small" : "large"); }
  return result; /* not sorted, but rows are unique */
}
